;;; the following code illustrates where clisp fails with
;;; symbol-macrolet and loop
(let ((max-iter 10)
      (i-iter 0))
  (symbol-macrolet 
      ((inc-iter (incf i-iter)))
    (loop
       (when (> i-iter max-iter)
	 (return i-iter))
  ;;;     inc-iter  ;;; this line will cause a compilation error
       (incf i-iter)
       (print i-iter))))


#|
;;; The reported error is
LOOP: illegal syntax near #1=(WHEN (> I-ITER MAX-ITER)
    (RETURN I-ITER)) in #2=(LOOP #1# INC-ITER (INCF I-ITER)
    (PRINT I-ITER))
   [Condition of type SYSTEM::SIMPLE-SOURCE-PROGRAM-ERROR]
|#

;;; the error is fixed with labels instead of symbol-macrolet
(let ((max-iter 10)
      (i-iter 0))
  (labels ((inc-iter ()
	     (incf i-iter)))
    (loop
       (when (> i-iter max-iter)
	 (return i-iter))
       (inc-iter) ;;; <-- fix
       (print i-iter))))


;; the modified macro with-ode-integration, but with the offending
;; make-next-step symbol macro replaced with a labels function
(defmacro with-ode-integration1
    ((function time step-size max-time
	       dependent dimensions &optional (stepper '*step-rk8pd*)
	       (absolute-error 1.0d-6) (relative-error 0.0d0))
     &body body)
  "Environment for integration of ordinary differential equations."
  (let ((dep (make-symbol "DEP"))
	(ctime (make-symbol "CTIME"))
	(cstep (make-symbol "CSTEP")))
    `(let ((stepperobj (make-ode-stepper ,stepper ,dimensions))
	   (control (make-y-control ,absolute-error ,relative-error))
	   (evolve (make-ode-evolution ,dimensions))
	   (,dep
	    (make-marray 'double-float :dimensions ,dimensions))
	   (,ctime (make-marray 'double-float :dimensions 1))
	   (,cstep (make-marray 'double-float :dimensions 1)))
       (labels ((make-next-step ()
		  (apply-evolution
		   evolve control stepperobj
		   ,function ,ctime
		   ,max-time ,cstep ,dep)))
       (symbol-macrolet
	   ((,time (maref ,ctime 0))
	    (,step-size (maref ,cstep 0))
	    ,@(loop for symb in dependent
		 for i from 0
		 collect `(,symb (maref ,dep ,i))))
	 ,@body)))))

;;;; test functions based on integrate-vanderpol.
;; the first one uses the modified macro
(defun integrate-vanderpol1
    (max-time &optional (step-size 1.0d-6) (stepper *step-rk8pd*) (print-steps t))
  "Integrate the van der Pol oscillator as given in Section 25.5 of the
   GSL manual.  To reproduce that example, (integrate-vanderpol 100.0d0)."
  (let ((mu 10.0d0) (initial-time 0.0d0) (iter 0))
    (declare (special mu))
    (with-ode-integration1
	(#-callback-toplevel-only
	 (make-ode-functions vanderpol vanderpol-jacobian 2)
	 #+callback-toplevel-only *vanderpol-cb*
	 time step max-time (dep0 dep1) 2 stepper)
      (setf dep0 1.0d0 dep1 0.0d0 step step-size time initial-time)
      (loop
	 (when (or (>= time max-time) (> iter *max-iter*))
	   (return (values iter time dep0 dep1)))
	 (make-next-step)
	 (incf iter)
	 (when print-steps
	   (format t "~12,6f~10t~12,6f~24t~12,6f~&" time dep0 dep1))))))

;; the second one uses gsll's macro, but I manually expanded
;; make-next-step and inserted it into the function body
(defun integrate-vanderpol2
    (max-time &optional (step-size 1.0d-6) (stepper *step-rk8pd*) (print-steps t))
  "Integrate the van der Pol oscillator as given in Section 25.5 of the
   GSL manual.  To reproduce that example, (integrate-vanderpol 100.0d0)."
  (let ((mu 10.0d0) (initial-time 0.0d0) (iter 0))
    (declare (special mu))
    (LET
	((STEPPEROBJ (MAKE-ODE-STEPPER STEPPER 2)) 
	 (CONTROL (MAKE-Y-CONTROL 1.0d-6 0.0d0))
	 (EVOLVE (MAKE-ODE-EVOLUTION 2)) 
	 (DEP (MAKE-MARRAY 'DOUBLE-FLOAT :DIMENSIONS 2))
	 (CTIME (MAKE-MARRAY 'DOUBLE-FLOAT :DIMENSIONS 1))
	 (CSTEP (MAKE-MARRAY 'DOUBLE-FLOAT :DIMENSIONS 1)))
      (SYMBOL-MACROLET
	  ((TIME (MAREF CTIME 0)) (STEP (MAREF CSTEP 0)) (DEP0 (MAREF DEP 0))
	   (DEP1 (MAREF DEP 1))
	   (MAKE-NEXT-STEP
	    (APPLY-EVOLUTION EVOLVE CONTROL STEPPEROBJ
			     (MAKE-ODE-FUNCTIONS VANDERPOL VANDERPOL-JACOBIAN 2) CTIME MAX-TIME CSTEP DEP)))
	(SETF DEP0 1.0d0 DEP1 0.0d0 STEP STEP-SIZE TIME INITIAL-TIME)
	(LOOP
	   (WHEN (OR (>= TIME MAX-TIME) (> ITER *MAX-ITER*)) 
	     (RETURN (VALUES ITER TIME DEP0 DEP1)))
	   ;; the following form replaces make-next-step
	   (APPLY-EVOLUTION EVOLVE CONTROL STEPPEROBJ
			    (MAKE-ODE-FUNCTIONS VANDERPOL VANDERPOL-JACOBIAN 2) 
			    CTIME MAX-TIME CSTEP DEP)
	   (INCF ITER)
	   (WHEN PRINT-STEPS 
	     (FORMAT T "~12,6f~10t~12,6f~24t~12,6f~&" TIME DEP0 DEP1)))))))

#|

;;;; test suite based on gsll's tests.  I compare the results of
;;;;integrate-vanderpol1 to integrate-vanderpol1

(defmacro build-test (stepper)
  "Comparison of two versions of vanderpol solver"
  (let ((stepper-label (onlisp:symb "*STEP-" stepper "*")))
    `(lisp-unit:define-test ,stepper
       (multiple-value-bind (dep0x dep1x)
	   (integrate-vanderpol1 1.0d0 1.d-4 ,stepper-label nil)
	 (multiple-value-bind (dep0t dep1t)
	     (integrate-vanderpol2 1.0d0 1.d-4 ,stepper-label nil)
	   (lisp-unit:assert-equal dep0x dep0t)
	   (lisp-unit:assert-equal dep1x dep1t))))))

;; a lisp master would have done a `dolist', but I could not figure
;; out the symbols
(progn
  "Build tests"
  (build-test RK2)
  (build-test RK4)
  (build-test RKF45)
  (build-test RKCK)
  (build-test RK8PD)
  (build-test RK2IMP)
  (build-test RK4IMP)
  (build-test BSIMP)
  (build-test GEAR1)
  (build-test GEAR2))

;; again, should have been done with a loop
(progn
  "Execute tests"
  (lisp-unit:run-tests RK2)
  (lisp-unit:run-tests RK4)
  (lisp-unit:run-tests RKF45)
  (lisp-unit:run-tests RKCK)
  (lisp-unit:run-tests RK8PD)
  (lisp-unit:run-tests RK2IMP)
  (lisp-unit:run-tests RK4IMP)
  (lisp-unit:run-tests BSIMP)
  (lisp-unit:run-tests GEAR1)
  (lisp-unit:run-tests GEAR2))
|#